from allauth.account.signals import user_logged_in
from allauth.socialaccount.adapter import DefaultSocialAccountAdapter, get_adapter
from allauth.socialaccount.providers import registry
from django.contrib import messages
from django.dispatch import receiver
from .socialaccount.providers.adfs_oauth2.provider import ADFSOAuth2Provider
from .socialaccount.adapter import SocialAccountAdapter

@receiver(user_logged_in)
def ensure_staff_login_via_adfs(**kwargs):
    adapter = get_adapter()
    if isinstance(adapter, SocialAccountAdapter):
        sociallogin = kwargs.get("sociallogin")
        via_adfs = sociallogin and sociallogin.account.provider == ADFSOAuth2Provider.id
        if not via_adfs:
            changed, user = adapter.update_user_fields(kwargs["request"], user=kwargs["user"])
            if changed:
                user.save()
                provider = registry.by_id(ADFSOAuth2Provider.id)
                messages.warning(kwargs["request"], 'User account modified due to log in provider. Log in with the %s provider to restore functionality when needed.' % provider.name)



from django.core.cache import DEFAULT_CACHE_ALIAS
try:
    from django.core.cache import caches
except ImportError:
    from django.core.cache import get_cache
    
    class CacheFallback(object):
        def __getitem__(self, alias):
            return get_cache(alias)
    
    caches = CacheFallback()


from django.core.exceptions import ImproperlyConfigured
from allauth.socialaccount import providers
from allauth.socialaccount.adapter import get_adapter
from allauth.socialaccount.providers.base import ProviderAccount
from allauth.socialaccount.providers.oauth2.provider import OAuth2Provider
from .utils import default_extract_extra_data_handler, default_extract_uid_handler, default_extract_common_fields_handler, default_extract_email_addresses_handler

class ADFSOAuth2Account(ProviderAccount):
    pass

class ADFSOAuth2Provider(OAuth2Provider):
    id = 'adfs_oauth2'
    name = 'ADFS Oauth2'
    package = 'allauth_adfs.socialaccount.providers.adfs_oauth2'
    account_class = ADFSOAuth2Account
    
    def get_auth_params(self, request, action):
        params = super(ADFSOAuth2Provider, self).get_auth_params(request, action)
        if "resource" not in params:
            raise ImproperlyConfigured("'resource' must be supplied as a key of the AUTH_PARAMS dict under adfs_oauth2 in the SOCIALACCOUNT_PROVIDERS setting.")
        return params
    
    def extract_extra_data(self, data):
        app = self.get_app(self.request)
        return self.get_settings().get("extract_extra_data_handler", default_extract_extra_data_handler)(data, app)
    
    def extract_uid(self, data):
        app = self.get_app(self.request)
        return self.get_settings().get("extract_uid_handler", default_extract_uid_handler)(data, app)

    def extract_common_fields(self, data):
        app = self.get_app(self.request)
        return self.get_settings().get("extract_common_fields_handler", default_extract_common_fields_handler)(data, app)
    
    def extract_email_addresses(self, data):
        app = self.get_app(self.request)
        return self.get_settings().get("extract_email_addresses_handler", default_extract_email_addresses_handler)(data, app)

provider_classes = [ADFSOAuth2Provider]



from six import text_type
from django.utils.encoding import force_bytes, force_text
from uuid import UUID
from struct import pack
from base64 import urlsafe_b64encode, urlsafe_b64decode
from allauth.account.models import EmailAddress

def decode_payload_segment(s):
    """
       reference:
           https://github.com/jpadilla/pyjwt/blob/528318787eff3df062f2b55a5f79964aece74f18/jwt/utils.py#L12 
    """
    if isinstance(s, text_type):
        s = s.encode('ascii')
    
    rem = len(s) % 4

    if rem > 0:
        s += b'=' * (4 - rem)

    return force_text(urlsafe_b64decode(s))

def parse_token_payload_segment(t):
    """
        reference:
            https://github.com/jpadilla/pyjwt/blob/4f899c6764d57000eba0fc40721f9e1b5d94a77a/jwt/api_jws.py#L130
    """
    t = force_bytes(t)
    try:
        signing_input, crypto_segment = t.rsplit(b'.', 1)
        header_segment, payload_segment = signing_input.split(b'.', 1)
    except ValueError:
        raise ValueError('Not enough segments')
    
    return payload_segment

def default_extract_uid_handler(data, app):
    guid = force_bytes(data['guid'])
    raw = urlsafe_b64decode(guid)
    uid = UUID(bytes_le=raw)
    return text_type(uid)

def per_social_app_extract_uid_handler(data, app):
    guid = force_bytes(data['guid'])
    raw = urlsafe_b64decode(guid)
    uid = UUID(bytes_le=raw)
    return "{};{}".format(app.id, uid)

def default_extract_common_fields_handler(data, app):
    upn = data['upn']
    common_fields = dict(
        username = upn.split("@")[0],
        first_name = data.get('first_name'),
        last_name = data.get('last_name'),
        email = data.get('email', upn),
    )
    for key in ("is_staff", "is_superuser", "is_active"):
        common_fields[key] = data.get(key) == "1"
    return common_fields

def per_social_app_extract_common_fields_handler(data, app):
    common_fields = default_extract_common_fields_handler(data, app)
    uid_bytes = UUID(default_extract_uid_handler(data, app)).bytes
    uid_b64 = urlsafe_b64encode(uid_bytes)
    aid_bytes = pack("I", app.id-1)# I format is 0-4294967295
    aid_b64 = urlsafe_b64encode(aid_bytes)
    username = "".join([uid_b64, aid_b64]).replace("=", "")# length of 28
    common_fields["username"] = username
    return common_fields

def default_extract_email_addresses_handler(data, app):
    addressess = []
    common_fields = default_extract_common_fields_handler(data, app)
    email = common_fields.get("email")
    if email:
        addressess.append(EmailAddress(email=email, verified=True, primary=True))
    return addressess

def default_extract_extra_data_handler(data, app):
    return data



from allauth.socialaccount.providers.oauth2.views import (OAuth2Adapter,
                                                          OAuth2LoginView,
                                                          OAuth2CallbackView)
from django.core.exceptions import ImproperlyConfigured
from django.utils.six import string_types
from .provider import ADFSOAuth2Provider
from .utils import decode_payload_segment, parse_token_payload_segment
import requests
from xml.dom.minidom import parseString
from hashlib import md5

try:
    from urllib.parse import urlunsplit
except ImportError:
    from urlparse import urlunsplit

try:
    import jwt
    from cryptography.x509 import load_der_x509_certificate
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization
except ImportError:
    JWT_AVAILABLE = False
    import json
else:
    JWT_AVAILABLE = True

from .compat import caches, DEFAULT_CACHE_ALIAS
    

class ADFSOAuth2Adapter(OAuth2Adapter):
    provider_id = ADFSOAuth2Provider.id
    
    def get_setting(self, key, default="", required=True):
        value = self.get_provider().get_settings().get(key, default)
        if not value and required:
            raise ImproperlyConfigured("ADFS OAuth2 provider setting '%s' is required.  It must not be falsey." % key)
        return value
    
    @property
    def redirect_uri_protocol(self):
        value = self.get_setting("redirect_uri_protocol", default=None, required=False)
        if isinstance(value, string_types):
            value = value.lower()
        if value not in ("http", "https", None):
            raise ImproperlyConfigured("ADFS OAuth2 provider setting 'redirect_uri_protocol' must be one of 'http', 'https', or None. You supplied '%s'." % value)
        return value
    
    @property
    def host(self):
        """
            e.g. sso.internal.example.com or sso.example.com:8443
        """
        return self.get_setting("host")
    
    def construct_adfs_url(self, path):
        parts = (
            "https",
            self.host,
            path,
            "",
            "",
        )
        return urlunsplit(parts)
    
    @property
    def access_token_url(self):
        return self.construct_adfs_url("/adfs/oauth2/token")
    
    @property
    def authorize_url(self):
        return self.construct_adfs_url("/adfs/oauth2/authorize")
    
    @property
    def federation_metadata_url(self):
        return self.construct_adfs_url("/FederationMetadata/2007-06/FederationMetadata.xml")
    
    @property
    def federation_metadata_xml(self):
        response = requests.get(self.federation_metadata_url)
        
        if response.status_code == 200:
            data = response.content
        else:
            raise RuntimeError("Could not retrieve federation metadata")
        
        xml = parseString(data)
        
        return xml
    
    @property
    def token_signature_key(self):
        cache_alias = self.get_setting("token_signature_key_cache_alias", DEFAULT_CACHE_ALIAS)
        cache = caches[cache_alias]
        cache_key = ":".join([
            "allauth_adfs",
            "ADFSOAuth2Adapter",
            md5(self.federation_metadata_url).hexdigest(),
            "token_signature_key",
        ])
        
        pub = cache.get(cache_key)
        
        if pub is None:
            xml = self.federation_metadata_xml
            
            signature = xml.getElementsByTagName("ds:Signature")[0]
            cert_b64 = signature.getElementsByTagName("X509Certificate")[0].firstChild.nodeValue
            
            cert_str = decode_payload_segment(cert_b64)
            cert_obj = load_der_x509_certificate(cert_str, default_backend())
            
            pub = cert_obj.public_key().public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
             
            timeout = self.get_setting("token_signature_key_cache_timeout", 0, required=False)
            cache.set(cache_key, pub, timeout)
        
        return pub

    def complete_login(self, request, app, token, **kwargs):
        verify_token = self.get_setting("verify_token", True, required=False)
        
        if JWT_AVAILABLE:
            kwargs = {"verify": verify_token}
            if verify_token:
                auth_params = self.get_setting("AUTH_PARAMS")
                
                try:
                    kwargs["audience"] = "microsoft:identityserver:%s" % auth_params["resource"]
                except KeyError:
                    raise ImproperlyConfigured("ADFS OAuth2 AUTH_PARAMS setting 'resource' must be specified.")
                
                kwargs["leeway"] = self.get_setting("time_validation_leeway", 0, required=False)
                
                kwargs["key"] = self.token_signature_key
                
            payload = jwt.decode(token.token, **kwargs)
            
        else:
            if verify_token:
                raise ImproperlyConfigured("ADFS OAuth2 cannot verify tokens without the `PyJWT` and `cryptography` packages.  They can both be installed with pip.  The `cryptography` package requires development headers for python and libffi.  They can be installed with 'apt-get install python-dev libffi-dev' on Ubuntu Linux.  You can disable token verification by setting 'verify_token' to False under the 'adfs_oauth2' socialaccount provider configuration dictionary in `settings.py`.  IT IS NOT RECOMMENDED TO DISABLE TOKEN VERIFICATION IN PRODUCTION!")
            
            encoded_data = parse_token_payload_segment(token.token)
            data = decode_payload_segment(encoded_data)
            payload = json.loads(data)
        
        return self.get_provider().sociallogin_from_response(
            request,
            payload
        )

oauth_login = OAuth2LoginView.adapter_view(ADFSOAuth2Adapter)
oauth_callback = OAuth2CallbackView.adapter_view(ADFSOAuth2Adapter)


from setuptools import setup, find_packages
from setuptools.command.test import test as SetuptoolsTestCommand
from shlex import split
from sys import version_info

class RunTestsCommand(SetuptoolsTestCommand):
    user_options = [
        ('only=', 'o', 'Only run the specified tests'),
        ('level=', 'l', 'Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output'),
        ('suppress-coverage-report', None, 'Suppress coverage report'),
    ]
    def initialize_options(self):
        SetuptoolsTestCommand.initialize_options(self)
        self.test_suite = "override"
        self.only = ""
        self.level = "1"
        self.suppress_coverage_report = None

    def finalize_options(self):
        SetuptoolsTestCommand.finalize_options(self)
        self.test_suite = None
        self.level = int(self.level)
        self.suppress_coverage_report = self.suppress_coverage_report is not None

    def run(self):
        SetuptoolsTestCommand.run(self)
        self.with_project_on_sys_path(self.run_tests)

    def run_tests(self):
        import coverage.cmdline
        import os
        import subprocess
        import sys
        import time
        
        owd = os.path.abspath(os.getcwd())
        nwd = os.path.abspath(os.path.dirname(__file__))
        os.chdir(nwd)
        tests = split(self.only)
        if not tests:
            tests.extend([nwd, os.path.abspath('test_project')])
        errno = coverage.cmdline.main(['run', os.path.abspath('test_project/manage.py'), 'test', '--verbosity=%d' % self.level] + tests)
        
        if not self.suppress_coverage_report:
            coverage.cmdline.main(['report', '-m'])
        
        if None not in [os.getenv("TRAVIS", None), os.getenv("TRAVIS_JOB_ID", None), os.getenv("TRAVIS_BRANCH", None)]:
            env = os.environ.copy()
            env["PYTHONPATH"] = os.pathsep.join(sys.path)
            cmd = ["coveralls"]
            coveralls_retry = 5
            while subprocess.call(cmd, env=env) and coveralls_retry:
                coveralls_retry -= 1
                if coveralls_retry:
                    seconds = 10
                    print("coveralls was unsuccessful. sleeping for %s seconds before retrying." % seconds)
                    time.sleep(seconds)
                else:
                    print("coveralls failed.")
        
        os.chdir(owd)
        
        raise SystemExit(errno)

pki_require = ["certifi"]
if version_info < (3, 0):
    pki_require = pki_require + ["pyopenssl", "ndg-httpsclient"]

tests_require = ['coverage', 'beautifulsoup4', 'html5lib', 'coveralls']
if version_info < (3, 3):
    tests_require = tests_require + ['mock', 'pbr<1.7.0']

setup(
    name = "django-allauth-adfs",
    version = "0.1.3",
    author = "gordon",
    author_email = "wgordonw1@gmail.com",
    description = "ADFS oAuth provider for django-allauth",
    url = "https://github.com/thenewguy/django-allauth-adfs",
    cmdclass={'test': RunTestsCommand},
    packages=find_packages(),
    extras_require={
        "jwt": ["PyJWT", "cryptography"],
        "pki": pki_require,
    },
    install_requires=['django-allauth>=0.26.0', 'six'],
    tests_require=tests_require,
    classifiers = [
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Framework :: Django',
    ],
)



import base64
import json
import six
import unittest

from allauth.socialaccount import providers
from allauth.socialaccount.models import SocialApp
from allauth.socialaccount.templatetags.socialaccount import get_providers
from allauth.socialaccount.tests import OAuth2TestsMixin
from allauth.tests import MockedResponse, TestCase
from django.conf import settings
from django.contrib.sites.models import Site
from django.template import RequestContext, Template
from django.test.client import RequestFactory
from django.urls import reverse

from .provider import ADFSOAuth2Provider
from .utils import decode_payload_segment, parse_token_payload_segment, default_extract_uid_handler


def encode(source):
    if six.PY3:
        source = source.encode('utf-8')
    content = base64.b64encode(source).decode('utf-8')
    return content.strip()


class TestProviderUrls(TestCase):
    def test_urls_importable(self):
        from allauth_adfs.socialaccount.providers.adfs_oauth2 import urls
    
    def test_urls_populated(self):
        from allauth_adfs.socialaccount.providers.adfs_oauth2 import urls
        self.assertIsInstance(urls.urlpatterns, list)
        self.assertTrue(urls.urlpatterns)
        
    def test_login_url(self):
        registry = providers.ProviderRegistry()
        registry.load()
        provider = registry.by_id(ADFSOAuth2Provider.id)
        login_url = provider.get_login_url(request=None)
        self.assertEquals(login_url, "/accounts/adfs_oauth2/login/")
    
    def test_template_login_url(self):
        registry = providers.ProviderRegistry()
        registry.load()
        provider = registry.by_id(ADFSOAuth2Provider.id)
        
        factory = RequestFactory()
        request = factory.get('/accounts/login/')
        c = RequestContext(request, {
            'provider': provider,
        })
        t = Template("""
            {% load socialaccount %}
            {% provider_login_url provider.id %}
        """)
        content = t.render(c).strip()
        
        self.assertEquals(content, "/accounts/adfs_oauth2/login/")


class TestProvidersRegistryFindsUs(TestCase):
    def test_load(self):
        registry = providers.ProviderRegistry()
        self.assertFalse(registry.loaded)
        self.assertFalse(registry.provider_map)
        self.assertNotIn(ADFSOAuth2Provider.id, registry.provider_map)
        registry.load()
        self.assertIn(ADFSOAuth2Provider.id, registry.provider_map)
        provider = registry.by_id(ADFSOAuth2Provider.id)
        self.assertIsInstance(provider, ADFSOAuth2Provider)


class UtilsTests(TestCase):
    def test_guid(self):
        data = {"guid": "2brp/e0eREqX7SzEA6JjJA=="}
        uid = default_extract_uid_handler(data, None)
        self.assertEquals(uid, six.text_type('fde9bad9-1eed-4a44-97ed-2cc403a26324'))
        


class ADFSTests(OAuth2TestsMixin, TestCase):
    provider_id = ADFSOAuth2Provider.id
    default_claims = {
        "guid": "2brp/e0eREqX7SzEA6JjJA==",
        "upn": "foo@bar.example.com",
        "first_name": "jane",
        "last_name": "doe"
    }
    
    def get_mocked_response(self):
        return MockedResponse(200, '')
    
    def get_login_response_json(self, **kwargs):
        jwt = self.get_dummy_jwt()
        return '{"access_token":"%s"}' % jwt
    
    def test_unencrypted_token_payload(self):
        jwt = self.get_dummy_jwt()
        
        encoded_claims_json = parse_token_payload_segment(jwt)
        decoded_claims_json = decode_payload_segment(encoded_claims_json)
        parsed_claims = json.loads(decoded_claims_json)
        
        claims = self.default_claims
        
        self.assertEqual(claims["guid"], parsed_claims["guid"])
        self.assertEqual(claims["upn"], parsed_claims["upn"])
        self.assertEqual(claims["first_name"], parsed_claims["first_name"])
        self.assertEqual(claims["last_name"], parsed_claims["last_name"])
    
    @unittest.skip("refresh tokens are not supported")
    def test_account_refresh_token_saved_next_login(self, **kwargs):
        pass
    
    @unittest.skip("cannot match expected token value")
    def test_account_tokens(self, **kwargs):
        pass
    
    def get_dummy_jwt(self, claims=None):
        if claims is None:
            claims = self.default_claims
        
        # raw data
        header = {
            "alg": "none",
            "typ":"JWT"
        }
        
        signature = ""
        
        # payload data
        header_data = encode(json.dumps(header))
        claims_data = encode(json.dumps(claims))
        signature_data = encode(signature)
        payload = [header_data, claims_data, signature_data]
        
        return ".".join(payload)



from allauth.account.signals import user_logged_in
from allauth.exceptions import ImmediateHttpResponse
from allauth.socialaccount.adapter import DefaultSocialAccountAdapter, get_adapter
from allauth.socialaccount.providers import registry
from django.contrib import messages
from django.dispatch import receiver
from django.http import HttpResponseForbidden
from .providers.adfs_oauth2.provider import ADFSOAuth2Provider

class SocialAccountAdapter(DefaultSocialAccountAdapter):
    def pre_social_login(self, request, sociallogin):
        # new user logins are handled by populate_user
        if sociallogin.is_existing:
            changed, user = self.update_user_fields(request, sociallogin)
            if changed:
                user.save()

    def populate_user(self, request, sociallogin, data):
        user = super(SocialAccountAdapter, self).populate_user(request, sociallogin, data)
        self.update_user_fields(request, sociallogin, user)
        return user
    
    def update_user_fields(self, request, sociallogin=None, user=None):
        changed = False
        if user is None:
            user = sociallogin.account.user
        adfs_provider = registry.by_id(ADFSOAuth2Provider.id, request)
        
        false_keys = ["is_staff", "is_superuser"]
        boolean_keys = false_keys + ["is_active"]
        copy_keys = boolean_keys + ["first_name", "last_name", "email"]
        
        if sociallogin is not None and sociallogin.account.provider == ADFSOAuth2Provider.id:
            data = sociallogin.account.extra_data
            values = adfs_provider.extract_common_fields(data)
            for key in copy_keys:
                # it is assumed that values are cleaned and set for all
                # fields and if any of the boolean_keys are not provided
                # in the raw data they should be set to False by
                # the extract_common_fields method
                if getattr(user, key) != values[key]:
                    setattr(user, key, values[key])
                    changed = True
        else:
            for key in false_keys:
                if getattr(user, key):
                    msg = "Staff users must authenticate via the %s provider!" % adfs_provider.name
                    response = HttpResponseForbidden(msg)
                    raise ImmediateHttpResponse(response)
        
        return changed, user
